Imports System.IO

Public Class Data

  ' Main Objects
  Public NumExams As Integer
  Public LExams() As LExam
  Public NumRooms As Integer
  Public LRooms() As LRoom
  Public NumTimeSlots As Integer
  Public LTimeSlots() As LTimeSlot

  ' Soft Penalties
  Public penTwoInARow As Integer = 0
  Public penTwoInADay As Integer = 0
  Public penWiderSpread As Integer = 0
  Public penDiffDurations As Integer = 0
  Public frontNumLargestExams As Integer = 0
  Public frontNumLastPeriods As Integer = 0
  Public penFrontLoading As Integer = 0

  ' Misc
  Public UNSCHEDULED_PENALTY As Integer = 5000 '555000

  ' Temp working objects and vars
  Public NumStudents As Integer = 0
  Public ExamStudents() As ArrayList    ' of studentIndex

  ' Algorithm Code
  Private bestEval As Long = 0
  Public bReporting As Boolean = False

  ' What to show
  Private bReportOnProblemData As Boolean = False
  Public bEvaluationSummary As Boolean = False
  Private bReportingHard As Boolean = True

  ' Don't need the full functinality of 'New' when we're running as a WebApp - the webpage will call all required methods itself
  Private bRunAsWebApplication As Boolean = False

  Public Sub New(ByVal filename As String, ByVal action As String)

    Dim bStillOkay As Boolean = True

    If Not (bRunAsWebApplication) Then
      Try

        ' Problem (exam) file
        bStillOkay = LoadData(filename)
        If Not (bStillOkay) Then
          Console.WriteLine("Cannot Read Problem File")
          Return
        End If

        ' Solution (sln) file
        bStillOkay = LoadSolution(filename)
        If Not (bStillOkay) Then
          Console.WriteLine("Cannot Read Solution File")
          Return
        End If

        ' Initialise Stuff
        Initialise()

        ' Report
        bEvaluationSummary = True
        If Not (action Is Nothing) Then
          ' Violations
          If action.ToLower = "violations" Then
            bReporting = True
          End If
        End If

        Evaluate()

      Catch ex As Exception
        Console.WriteLine("Problem with Validator: " & ex.Message)
      End Try

    End If

  End Sub

  Private Sub Initialise()

    ' Mark largest X events and last Y periods (frontload)
    ' Exams - Collect durations and indices, sort, then mark the first X of them
    ' Collect
    Dim ei As Integer, i As Integer, pi As Integer, durExams As New ArrayList ' of IndexValue
    For ei = 0 To NumExams - 1
      durExams.Add(New IndexValue(ei, LExams(ei).ClassSize))
    Next

    ' Sort
    durExams.Sort(New IndexValue.ivSortMethodDouble)      ' Sort descending - largest first
    'Dim str As String
    'For i = 0 To durExams.Count - 1
    '  str &= CType(durExams(i), IndexValue).Value & " "
    'Next
    'Console.WriteLine(str)

    ' Now mark
    For i = 0 To frontNumLargestExams - 1
      ei = CType(durExams(i), IndexValue).Index
      If ei >= 0 And ei < NumExams Then
        LExams(ei).bLargestExams = True
      End If
    Next
    ' Periods - simply mark the last Y periods
    For pi = NumTimeSlots - frontNumLastPeriods To NumTimeSlots - 1
      If pi >= 0 And pi < NumTimeSlots Then
        LTimeSlots(pi).bLastPeriods = True
      End If
    Next

  End Sub


  Public Sub GoSchedule(ByVal filename As String, ByVal numCon As Integer, ByVal numImp As Integer)
    If Not (bRunAsWebApplication) Then
      Try

        '''''''''''''''''''''
        ' Initialisation
        '''''''''''''''''''''

        ' Clear all Room/Period events
        Dim ri As Integer, pi As Integer
        For ri = 0 To NumRooms - 1
          For pi = 0 To NumTimeSlots - 1
            LRooms(ri).Events(pi) = New ArrayList
          Next
        Next

        ' Auto-Schedule
        '   If action = "schedule" Then
        bReporting = False
        bReportingHard = False
        bEvaluationSummary = False
        Adaptive(numCon)
        GreatDeluge(numImp, 1.1)
        '   Console.WriteLine("AFTER DELUGE")
        bReporting = False 'True
        bEvaluationSummary = True
        bReportingHard = True
        Evaluate()
        '   Console.WriteLine("AFTER LAST EVALUATE")
        SaveSolution(filename)
        '  Console.WriteLine("AFTER SAVE SOLUTION")
        '  End If
      Catch ex As Exception
        Console.WriteLine("Problem with Validator: " & ex.Message)
      End Try
    End If
  End Sub

#Region "Algorithms"

  Private Function Adaptive(ByVal LNumIterations As Integer) As Long

    Console.WriteLine()
    Console.WriteLine("Construction (Adaptive)")
    Console.WriteLine("=======================")
    Console.WriteLine()

    Dim i As Integer, j As Integer, ei As Integer, di As Integer, ti As Integer, it As Integer, ri As Integer, dei As Integer
    Dim bestFit As Long, pos As LPos, pen As Long, evalPen As Long, bestPos As LPos, k As Integer
    Dim numUnscheduled As Integer
    Dim eventArray As ArrayList
    Dim numResultsRequired As Integer = 10

    '  bCanFinish = False
    bestEval = Long.MaxValue
    ' Set up ordered list
    eventArray = GetEventOrderedList()

    Dim numUnscheduledBest As Integer = NumExams

    ' Begin loop
    For it = 0 To LNumIterations - 1

      ' Sort ordering
      eventArray.Sort(New IndexValue.ivSortMethodDouble)
      ' Clear
      ClearSchedule()
      numUnscheduled = NumExams

      evalPen = 0
      Dim deltaPen As Long = 0

      For i = 0 To NumExams - 1
        ei = CType(eventArray(i), IndexValue).Index

        ' Only do this if not already scheduled (ie. through a RecursiveSchedule, or fixed)
        If LExams(ei).pos.pi = -1 Then

          '  ' Already-specified Timeslot constraint - Just use the Existing position
          '  bestPos = New LPos(LExams(ei).pos.pi, -1)
          '  bestFit = PenaltyAtRecursive(bestPos, ei)

          'Else

          bestFit = Long.MaxValue
          pos = New LPos
          bestPos = New LPos

          For k = 0 To LExams(ei).SuitablePeriods.Count - 1
            ti = CInt(LExams(ei).SuitablePeriods(k))

            pos.pi = ti
            pos.ri = -1
            '            pos.ClearRoomAllocation()

            If CanScheduleRecursive(pos, ei) Then

              pen = PenaltyAtRecursive(pos, ei)
              If pen <= bestFit Then
                ' Add to 
                bestFit = pen
                bestPos.pi = ti

                '' Won't find anything better than 0 penalty
                If bestFit = 0 Then
                  Exit For
                End If

              End If
            End If  ' CanScheduleEvent
          Next

          '   End If

          ' If a position was found
          If bestPos.pi <> -1 Then
            ' Found a bestpos
            numUnscheduled -= 1

            '      bestPos.ClearRoomAllocation()  ' Find best room
            bestPos.ri = -1
            ScheduleRecursive(bestPos, ei)
            ' If penalty, increase the sort value
            CType(eventArray(i), IndexValue).Value += bestFit
          Else
            ' Unplaced
            CType(eventArray(i), IndexValue).Value += 2880 '12880 '(1400 * LExams(ei).ConflictExams.Count) ' 2880 '1230
          End If


        Else
          ' This was already found scheduled - reduce the unscheduled count
          numUnscheduled -= 1
        End If

      Next  ' Event

      evalPen = Evaluate()
      If evalPen < bestEval Then
        ' Store the best Events pos details
        RecordBestSolution()
        bestEval = evalPen
        numUnscheduledBest = numUnscheduled
      End If


      If it Mod 5 = 0 Then
        Console.WriteLine("it: " & it + 1 & ", Eval: " & evalPen & ", Unsch:" & numUnscheduled & "  Best Eval: " & bestEval & "  Unsch:" & numUnscheduledBest) '& " (" & strUnscheduled & " )")
      End If

      If evalPen = 0 Then 'Or bCanFinish Then
        Exit For
      End If

    Next ' Iteration

    ' Restore the best Events pos details
    RestoreBestSolution()
    ' Final check on how many unscheduled
    numUnscheduled = 0
    For ei = 0 To NumExams - 1
      If LExams(ei).pos.pi = -1 Then
        numUnscheduled += 1
      End If
    Next
    'If numUnscheduled > 0 Then
    '  MsgBox("Non-feasible solution")
    'End If

    Console.WriteLine("it: " & it + 1 & ", Eval: " & evalPen & ", Unsch:" & numUnscheduled & "  Best Eval: " & bestEval & "  Unsch:" & numUnscheduledBest)

    Return Evaluate()

  End Function

  Private Function GetEventOrderedList() As ArrayList
    Dim eventArray As New ArrayList
    Dim ei As Integer
    For ei = 0 To NumExams - 1
      eventArray.Add(New IndexValue(ei, LExams(ei).ClassSize * LExams(ei).ConflictExams.Count))
    Next
    eventArray.Sort(New IndexValue.ivSortMethodDouble)
    Return eventArray
  End Function

  Private ImpHeatFrequency As Long = 10000
  Private InitialDecayRate As Double = 200  '50 '200
  Private ThenDecayRate As Double = 50  '2
  Private MoveSwapRatio As Integer = 3

  Private Function GreatDeluge(ByVal totalMoves As Long, ByVal multiple As Double) As Long

    Console.WriteLine()
    Console.WriteLine("Improvement (Great Deluge)")
    Console.WriteLine("==========================")
    Console.WriteLine()

    ImpHeatFrequency = 10000  'Int(totalMoves / 2)

    ' options
    '   Dim bTimeloop As Boolean = True, timeCount As Integer = 40 '100

    ' Dim timeDiff As Integer
    Dim currentEval As Long, deltaPen As Long, newEval As Long
    Dim boundary As Double, decayRate As Double, sinceImp As Integer ', strInfo As String
    Dim ei As Integer, dei As Integer, i As Integer ', timeBounds As Integer
    Dim numHits As Integer = 0
    Dim it As Integer
    Const InfoUpdateFactor = 1000 '50000

    Dim bTryToPlaceUnscheduled As Boolean = False     ' Gives more priority to those unscheduled events than improving already scheduled

    currentEval = Evaluate()

    bestEval = currentEval
    RecordBestSolution()

    boundary = CDbl(currentEval * multiple)
    sinceImp = 0
    decayRate = (boundary / CDbl(totalMoves) * InitialDecayRate)

    Dim newPos As LPos, oldPos As LPos

    ' Build list of unscheduled events
    Dim unEvents As New ArrayList
    For ei = 0 To NumExams - 1
      If LExams(ei).pos.pi = -1 Then
        unEvents.Add(ei)
      End If
    Next
    unEvents.Sort()

    it = 0
    While it < totalMoves 'numHits < totalMoves

      If it Mod InfoUpdateFactor = 0 Then
        Console.WriteLine("It: " & it & " BEST: " & bestEval & ", Curr: " & currentEval & ", numHits " & numHits & "  Unsch:" & unEvents.Count) ' & ", " & olddiffpen & ", " & newdiffpen)  '& " sinceimp=" & sinceImp) ' & ", took " & timeDiff & "s - score: " & strmore)
      End If

      sinceImp += 1
      ' Console.WriteLine("SinceImp " & sinceImp)
      '    numMoves += 1
      If boundary > 0 Then
        boundary -= decayRate
      End If

      If sinceImp > ImpHeatFrequency Then 'totalMoves / ScheduleProperties.ImpHeatFrequency Then
        boundary = CDbl(bestEval * multiple)
        If boundary < 10 Then
          boundary = 10
        End If
        sinceImp = 0
        decayRate = (boundary / CDbl(totalMoves - numHits) * ThenDecayRate) 'ScheduleProperties.ImpHeatFrequency)
        'Console.WriteLine()
        'Console.WriteLine()
        'Console.WriteLine("Boundary now " & boundary)
        'Console.WriteLine()
        'Console.WriteLine()
      End If

      it += 1

      ' Move
      Dim bValidMove As Boolean = False, ti As Integer, k As Integer

      If it Mod MoveSwapRatio > 0 Then

        If bTryToPlaceUnscheduled Then
          unEvents = New ArrayList
          For ei = 0 To NumExams - 1
            If LExams(ei).pos.pi = -1 Then
              unEvents.Add(ei)
            End If
          Next
          unEvents.Sort()
          Dim giveUpCount As Integer = 10 '100
          While bValidMove = False And giveUpCount > 0
            If unEvents.Count > 0 And giveUpCount > 0 Then
              k = CInt((unEvents.Count - 1) * Rnd())
              ei = CInt(unEvents(k))
              giveUpCount -= 1
            End If
            ' Pick one timeslot and try it
            For i = 0 To 5 '0 '5
              k = CInt((LExams(ei).SuitablePeriods.Count - 1) * Rnd())
              ti = CInt(LExams(ei).SuitablePeriods(k))
              newPos = New LPos(ti, -1)
              If CanSchedule(ei, newPos) Then
                bValidMove = True
                Exit For
              End If
            Next
          End While
        End If

        'While Not (bValidMove)
        '  If Not (bValidMove) Then
        ei = CInt((NumExams - 1) * Rnd())
        oldPos = New LPos(LExams(ei).pos.pi, LExams(ei).pos.ri)
        ' Pick one timeslot and try it
        k = CInt((LExams(ei).SuitablePeriods.Count - 1) * Rnd())
        ti = CInt(LExams(ei).SuitablePeriods(k))
        newPos = New LPos(ti, -1)
        If CanSchedule(ei, newPos) Then
          bValidMove = True
        End If
        '  End If
        'End While

        If bValidMove Then
          ' Make the move
          numHits += 1

          Unschedule(ei)
          Schedule(ei, newPos)

          newEval = Evaluate()
          deltaPen = newEval - currentEval

          'penOld = PenaltyAt(Data.LEvent(ei).Pos, ei)
          '' Need to schedule in newPos first to get the actual newPos room allocation to enable correct Penalty eval
          'penNew = PenaltyAt(newPos, ei)
          '' Work out difference and new overall evaluation
          'deltaPen = penNew - penOld

          ' Keep move if acceptable 
          If (deltaPen <= 0 Or (currentEval + deltaPen) <= boundary) Then
            currentEval = newEval

            ' Only reheat if there's been no improvement to current - not just besteval
            If deltaPen < 0 Then
              sinceImp = 0
            End If

            ' Did we schedule a previously unscheduled one?
            If unEvents.BinarySearch(ei) >= 0 Then
              unEvents.Remove(ei)
            End If

            '' Quick Delta evaluation check
            'If Evaluate() <> currentEval Then
            '  MsgBox("Problem - delta=" & currentEval & ", Absolute=" & Evaluate())
            'End If
          Else
            ' Move rejected - Put the events back in their original place
            Unschedule(ei)
            If Not (oldPos Is Nothing) Then
              Schedule(ei, oldPos)
            End If
          End If
        End If

      Else

        ' Swap
        Dim eiRoomIndexOld As Integer, deiRoomIndexOld As Integer, eiRoomIndexNew As Integer, deiRoomIndexNew As Integer
        Dim bValidSwap As Boolean = False

        '   Console.WriteLine("IN SWAP")

        While bValidSwap = False

          ei = CInt((NumExams - 1) * Rnd())
          dei = CInt((NumExams - 1) * Rnd())

          ' Not same event, and both are placed
          ti = LExams(ei).pos.pi
          Dim ti2 As Integer = LExams(dei).pos.pi
          If ei <> dei And ti <> -1 And ti2 <> -1 And _
            LExams(ei).SuitablePeriods.Contains(ti2) And LExams(dei).SuitablePeriods.Contains(ti) Then

            newPos = New LPos(ti2, LExams(dei).pos.ri)
            oldPos = New LPos(ti, LExams(ei).pos.ri)

            ' Need to unschedule first, to avoid conflicts
            Unschedule(ei)
            Unschedule(dei)

            ' Are times okay to swap?
            If CanSchedule(ei, newPos) Then
              ScheduleRecursive(newPos, ei)
              If CanSchedule(dei, oldPos) Then
                ScheduleRecursive(oldPos, dei)
                bValidSwap = True
              Else
                Unschedule(ei)
              End If
            End If

            ' Put back if not swapping
            If Not (bValidSwap) Then
              ' No valid swap was found - replace events
              ScheduleRecursive(oldPos, ei)
              ScheduleRecursive(newPos, dei)
            End If

          End If

        End While

        '      Console.WriteLine("OUT OF SWAP")

        If bValidSwap Then
          ' Make the move
          numHits += 1

          newEval = Evaluate()
          deltaPen = newEval - currentEval

          ' Keep swap if acceptable 
          If (deltaPen <= 0 Or (currentEval + deltaPen) <= boundary) Then
            currentEval = newEval

            ' Only reheat if there's been no improvement to current - not just besteval
            If deltaPen < 0 Then
              sinceImp = 0
            End If

            '' Quick Delta evaluation check (CourseBenchmark)
            'If EvaluateCourseBenchmark() <> currentEval Then
            '  MsgBox("Problem - delta=" & currentEval & ", Absolute=" & EvaluateCourseBenchmark())
            'End If

            '' Quick Delta evaluation check
            'If Evaluate() <> currentEval Then
            '  MsgBox("Problem - delta=" & currentEval & ", Absolute=" & Evaluate())
            'End If
          Else
            ' Swap rejected - Put the events back in their original place
            Unschedule(ei)
            Unschedule(dei)
            ScheduleRecursive(oldPos, ei)
            ScheduleRecursive(newPos, dei)
          End If
        End If

      End If


      ' Move / Swap done - check for best
      If currentEval < bestEval Then
        bestEval = currentEval
        RecordBestSolution()
        '   Console.WriteLine("Best")
        '  sinceImp = 0
      End If

      ' Can we Finish?
      If currentEval <= 0 Then 'Or bCanFinish Then
        Exit While
      End If

    End While

    ' Restore the best Events pos details
    RestoreBestSolution()

    ' Quick Delta evaluation check
    If Evaluate() <> bestEval Then
      '      MsgBox("Problem - delta=" & bestEval & ", Absolute=" & Evaluate())
      Console.WriteLine("Problem - delta=" & bestEval & ", Absolute=" & Evaluate())
    End If

    ' Console.WriteLine("It: " & it & " BEST: " & bestEval & ", Curr: " & currentEval & ", numHits " & numHits) ' & ", took " & timeDiff & "s - score: " & strmore)
    Console.WriteLine("It: " & it & " BEST: " & bestEval & ", Curr: " & currentEval & ", numHits " & numHits & "  Unsch:" & unEvents.Count) ' & ", took " & timeDiff & "s - score: " & strmore)

    Return Evaluate()
  End Function


  Private bestEvents As New ArrayList

  Private Sub RecordBestSolution()
    Dim ei As Integer
    bestEvents = New ArrayList
    For ei = 0 To NumExams - 1
      With LExams(ei).pos
        bestEvents.Add(New LPos(.pi, .ri))
      End With
    Next
  End Sub

  Private Sub RestoreBestSolution()
    Dim ei As Integer
    ClearSchedule()
    For ei = 0 To NumExams - 1
      Schedule(ei, CType(bestEvents(ei), LPos))
    Next
  End Sub

#End Region

#Region "Schedule Code"

  Private Function CanScheduleRecursive(ByVal pos As LPos, ByVal ei As Integer) As Boolean
    ' Return CanSchedule(ei, pos)

    If Not (CanSchedule(ei, pos)) Then
      Return False
    End If

    ' Now any coincidences
    Dim tmpEvents As New ArrayList, bCanSchedule As Boolean = True
    Dim i As Integer, dei As Integer, pos2 As New LPos(pos.pi, -1) ', sameRoom As Boolean
    '  pos.ClearRoomAllocation() '.RoomIndex2 = -1    ' reset to unallocated, so Schedule can pick best one
    Schedule(ei, pos)
    tmpEvents.Add(ei)

    If LExams(ei).coincidenceExams.Count > 0 Then
      For i = 0 To LExams(ei).coincidenceExams.Count - 1
        dei = CInt(LExams(ei).coincidenceExams(i))
        '   sameRoom = CBool(LEvent(ei).sameroom(i))
        ' Before checking, make sure pos2.timeslot is a suitable period for event dei
        If LExams(dei).SuitablePeriods.Contains(pos2.pi) Then
          ' Use same room as pos if asked for
          'If sameRoom Then
          '  pos2.RoomIndex2 = pos.RoomIndex2
          '  pos2.Size = LEvent(dei).students.Count
          'Else
          '  pos2.ClearRoomAllocation()
          'End If
          If (CanSchedule(dei, pos2)) Then
            ' Clear room allocation (which could be set on the CanScheduleEvent) if we don't have a 'sameroom' cooincidence
            'If Not sameRoom Then
            '  pos2.ClearRoomAllocation()
            'End If
            Schedule(dei, pos2)
            tmpEvents.Add(dei)
          Else
            bCanSchedule = False
            Exit For
          End If
        Else
          bCanSchedule = False
          Exit For
        End If
      Next
    End If
    ' Unschedule the temporarily scheduled events
    For i = 0 To tmpEvents.Count - 1
      Unschedule(CInt(tmpEvents(i)))
    Next
    Return bCanSchedule

  End Function

  Private Function CanSchedule(ByVal ei As Integer, ByVal pos As LPos) As Boolean
    Dim dei As Integer, i As Integer
    ' Basic Hard constraints
    ' Throw out any invalid schedules
    If pos.pi < 0 Or pos.pi >= NumTimeSlots Then
      Return False
    End If
    ' Can the timeslot accomodate the exam duration?
    '  Console.WriteLine(LExams(ei).Duration & ", " & LTimeSlots(pos.pi).Duration)
    If LExams(ei).Duration > LTimeSlots(pos.pi).Duration Then
      If bReportingHard Then
        Write(" Timeslot " & pos.pi & " (" & LTimeSlots(pos.pi).Duration & ") too short for Exam " & ei & " (" & LExams(ei).Duration & ") ")
      End If
      Return False
    End If
    ' Does exam conflict at this time - other scheduled conflict exams at this time?
    For i = 0 To LExams(ei).ConflictExams.Count - 1
      dei = CType(LExams(ei).ConflictExams(i), IndexValue).Index
      If LExams(dei).pos.pi = pos.pi Then
        ' Conflict exam in same timeslot - not good
        If bReportingHard Then
          Write(" Conflict at that time ")
        End If
        Return False
      End If
    Next
    ' Cooincidence
    For i = 0 To LExams(ei).coincidenceExams.Count - 1
      dei = CInt(LExams(ei).coincidenceExams(i))
      If LExams(dei).pos.pi >= 0 And LExams(dei).pos.pi < NumTimeSlots And LExams(dei).pos.pi <> pos.pi Then
        ' Exam has a cooincidence in some other location - this pi not good
        If bReportingHard Then
          Write(" Cooincidence with another exam in a different timeslot ")
        End If
        Return False
      End If
    Next
    ' Excludes
    For i = 0 To LExams(ei).excludesExams.Count - 1
      dei = CInt(LExams(ei).excludesExams(i))
      If LExams(dei).pos.pi >= 0 And LExams(dei).pos.pi < NumTimeSlots And LExams(dei).pos.pi = pos.pi Then
        ' Exam has an excludes with an exam at this location - this pi not good
        If bReportingHard Then
          Write(" Excludes with another exam in this timeslot ")
        End If
        Return False
      End If
    Next
    ' Before
    For i = 0 To LExams(ei).beforeExams.Count - 1
      dei = CInt(LExams(ei).beforeExams(i))
      If LExams(dei).pos.pi >= 0 And LExams(dei).pos.pi < NumTimeSlots And LExams(dei).pos.pi > pos.pi Then
        ' Exam has a before with an exam which comes after - this pi not good
        If bReportingHard Then
          Write(" Before with another exam in a later timeslot ")
        End If
        Return False
      End If
    Next
    ' After
    For i = 0 To LExams(ei).afterExams.Count - 1
      dei = CInt(LExams(ei).afterExams(i))
      If LExams(dei).pos.pi >= 0 And LExams(dei).pos.pi < NumTimeSlots And LExams(dei).pos.pi < pos.pi Then
        ' Exam has an after with an exam which comes before - this pi not good
        If bReportingHard Then
          Write(" After with another exam in an earlier timeslot ")
        End If
        Return False
      End If
    Next

    ' Exlusive Room
    ' If usage is > 0 of the room at this time, then it's being used already - if we need an exclusive room, then can't schedule there
    ' Also, if room is being used by an event which must be kept exclusive, we also can't schedule in it
    ' This is performed in FindRoom

    ' Find a Room
    If pos.ri < 0 Or pos.ri >= NumRooms Then
      If Not (CanAllocateRoom(ei, pos)) Then
        Return False
      End If
    End If

    ' Got here - okay
    Return True
  End Function

  Private Function CanAllocateRoom(ByVal ei As Integer, ByVal pos As LPos) As Boolean
    Dim ri As Integer = FindRoom(ei, pos)
    Return (ri >= 0 And ri < NumRooms)
  End Function

  'Private Function CanAllocateRoom(ByVal ei As Integer, ByVal pos As LPos) As Boolean
  '  Dim i As Integer, ri As Integer, k As Integer

  '  For k = 0 To LExams(ei).SuitableRooms.Count - 1 'NumRooms - 1
  '    ' Can the room accomodate the class size?
  '    ri = CInt(LExams(ei).SuitableRooms(k))
  '    ' Need to determine how much of the room is already being used at this time
  '    ' Crude - improve later
  '    Dim usage As Integer = 0
  '    For i = 0 To NumExams - 1
  '      If LExams(i).pos.pi = pos.pi And LExams(i).pos.ri = ri Then 'pos.ri Then
  '        ' Same time, same room - add size to the usage count
  '        usage += LExams(i).ClassSize
  '      End If
  '    Next
  '    If LExams(ei).ClassSize > LRooms(ri).Capacity - usage Then
  '      If bReporting Then
  '        Console.Write(" Room " & ri & " (" & LRooms(ri).Capacity - usage & " of " & LRooms(ri).Capacity & ") too small for Exam " & ei & " (" & LExams(ei).ClassSize & ") ")
  '      End If
  '      ' Return False
  '    Else
  '      Return True
  '    End If
  '  Next

  '  Return False
  'End Function

  Private Function FindRoom(ByVal ei As Integer, ByVal pos As LPos) As Integer
    Dim i As Integer, ri As Integer, k As Integer, dei As Integer, j As Integer

    For k = 0 To LExams(ei).SuitableRooms.Count - 1 'NumRooms - 1

      j = CInt((LExams(ei).SuitableRooms.Count - 1) * Rnd())
      ri = CInt(LExams(ei).SuitableRooms(j))

      '  ri = CInt(LExams(ei).SuitableRooms(k))

      ' Can the room accomodate the class size (can't use if class size + space already used in room > overall capacity?)
      If LExams(ei).ClassSize + LRooms(ri).Usage(pos.pi) > LRooms(ri).Capacity Then
        If bReportingHard Then
          Write(" Room " & ri & " (" & LRooms(ri).Capacity - LRooms(ri).Usage(pos.pi) & " left of " & LRooms(ri).Capacity & ") too small for Exam " & ei & " (" & LExams(ei).ClassSize & ") ")
        End If
        ' Does this exam require exclusive use of a room?
      ElseIf LExams(ei).bExclusiveRoom And LRooms(ri).Usage(pos.pi) > 0 Then
        ' Room already being used at this time (even though there's room)
        If bReportingHard Then
          Write(" Room " & ri & " is already being used, and Exam " & ei & " requires exclusive access")
        End If
        ' Are we putting the exam in a room which already has an 'exclusive-access' exam there at this time?
        ' Which exam is there?
      ElseIf Not (LRooms(ri).Events(pos.pi) Is Nothing) Then
        ' Check whether the exam in this room is an exclusive-access one
        Dim bFound As Boolean = False
        For i = 0 To LRooms(ri).Events(pos.pi).Count - 1
          dei = CInt(LRooms(ri).Events(pos.pi)(i))
          If LExams(dei).bExclusiveRoom Then
            bFound = True
            Exit For
          End If
        Next
        If bFound Then
          ' Room no good
          If bReportingHard Then
            WriteLine(" Room " & ri & " has an Exam " & dei & " at this time which requires exclusive access - can't put exam " & ei & " there")
          End If
        Else
          Return ri
        End If
      Else
        Return ri
      End If
    Next

    Return -1
  End Function

  Dim numTwoInARow As Integer, numTwoInADay As Integer, numWiderSpreads As Integer, numPeriodPenalties As Integer, numRoomPenalties As Integer
  Dim numMixDurPenalties As Integer, numFrontLoadPenalties As Integer

  Dim totpenTwoInADay As Integer
  Dim totpenTwoInARow As Integer
  Dim totpenWiderSpread As Integer
  Dim totpenPeriodPenalty As Integer
  Dim totpenRoomPenalty As Integer
  Dim totpenMixDurPenalty As Integer
  Dim totpenFrontLoadPenalty As Integer

  Dim strUnscheduled As String

  Public Function Evaluate() As Long
    Dim ei As Integer, pen As Long = 0, numun As Integer = 0
    numTwoInARow = 0
    numTwoInADay = 0
    numWiderSpreads = 0
    numPeriodPenalties = 0
    numRoomPenalties = 0
    numMixDurPenalties = 0
    numFrontLoadPenalties = 0
    totpenTwoInADay = 0
    totpenTwoInARow = 0
    totpenWiderSpread = 0
    totpenPeriodPenalty = 0
    totpenRoomPenalty = 0
    totpenMixDurPenalty = 0
    totpenFrontLoadPenalty = 0
    strUnscheduled = ""
    If bReporting Then
      WriteLine()
      WriteLine("Soft Constraint Violations")
      WriteLine("==========================")
    End If

    ' Exam penalties
    For ei = 0 To NumExams - 1
      If LExams(ei).pos.pi = -1 Or LExams(ei).pos.ri = -1 Then
        pen += UNSCHEDULED_PENALTY
        numun += 1
        strUnscheduled &= " " & CStr(ei)
        WriteLine("Exam " & ei & ": Unscheduled - Penalty " & UNSCHEDULED_PENALTY)
      Else
        ' What's the penalty for placement here
        pen += PenaltyForward(ei) ', LExams(ei).pos)
      End If
    Next

    ' MixDuration Penalties
    pen += AbsoluteDiffDurationsPenalty()

    If bEvaluationSummary Then
      WriteLine()
      WriteLine("Summary Evaluation")
      WriteLine("==================")
      WriteLine()
      WriteLine("Unscheduled: " & numun & " (of " & NumExams & "), pen=" & numun * UNSCHEDULED_PENALTY)
      WriteLine()
      WriteLine("TwoInARow: " & numTwoInARow & ", pen=" & totpenTwoInARow)
      WriteLine("TwoInADay: " & numTwoInADay & ", pen=" & totpenTwoInADay)
      WriteLine("WiderSpreads: " & numWiderSpreads & ", pen=" & totpenWiderSpread)
      WriteLine("PeriodPenalties: " & numPeriodPenalties & ", pen=" & totpenPeriodPenalty)
      WriteLine("RoomPenalties: " & numRoomPenalties & ", pen=" & totpenRoomPenalty)
      WriteLine("MixDurationPenalties: " & numMixDurPenalties & ", pen=" & totpenMixDurPenalty)
      WriteLine("FrontLoadPenalties: " & numFrontLoadPenalties & ", pen=" & totpenFrontLoadPenalty)
      WriteLine()
      WriteLine("Overall penalty = " & pen)
      If numun > 0 Then
        WriteLine()
        WriteLine("Infeasible Solution")
      End If
    End If
    Return pen
  End Function

  Private Function PenaltyForward(ByVal ei As Integer) As Long
    Return Penalty(ei, LExams(ei).pos, True)
  End Function

  Private Function PenaltyAtRecursive(ByVal pos As LPos, ByVal ei As Integer, Optional ByVal bForwardOnly As Boolean = False) As Long
    '  Return Penalty(ei, pos)
    Dim pen As Long = Penalty(ei, pos, bForwardOnly) ', penTextArray)
    ' Coincidences
    If LExams(ei).coincidenceExams.Count > 0 Then
      Dim i As Integer, dei As Integer
      For i = 0 To LExams(ei).coincidenceExams.Count - 1
        dei = CInt(LExams(ei).coincidenceExams(i))
        pen += Penalty(dei, pos, bForwardOnly) ', penTextArray)
      Next
    End If
    Return pen
  End Function

  Dim numExamConflicts As Integer = 0

  Private Function Penalty(ByVal ei As Integer, ByVal pos As LPos, Optional ByVal bForwardOnly As Boolean = False) As Long
    Dim pen As Long = 0, numConflicts As Integer = 0, i As Integer, penDiff As Long

    ' Two in a Row Constraints

    'WriteLine("")
    'WriteLine("TWOINROW")
    'WriteLine("========")

    ' Before (same day)
    numExamConflicts = 0
    If Not (bForwardOnly) And pos.pi > 0 Then
      If LTimeSlots(pos.pi - 1).lDate = LTimeSlots(pos.pi).lDate Then
        numConflicts += NumStudentConflicts(ei, pos.pi - 1)
      End If
    End If
    ' After (same day)
    If pos.pi < NumTimeSlots - 1 Then
      If LTimeSlots(pos.pi + 1).lDate = LTimeSlots(pos.pi).lDate Then
        numConflicts += NumStudentConflicts(ei, pos.pi + 1)
      End If
    End If
    'If numConflicts > 0 Then
    '  numTwoInARow += 1
    'End If
    If numExamConflicts > 0 Then
      numTwoInARow += numExamConflicts
    End If
    penDiff = (penTwoInARow * numConflicts)
    totpenTwoInARow += penDiff
    pen += penDiff
    If bReporting And penDiff > 0 Then
      WriteLine("Exam " & ei & ": TwoInARow Violation with " & numConflicts & " students - Penalty " & penDiff)
    End If

    'WriteLine("TWOINROW")
    'WriteLine("========")
    'WriteLine("")


    ' Two in a Day Constraints

    ' Same day (but not 2 in a row)
    numConflicts = 0
    numExamConflicts = 0
    ' Before (same day)
    If Not (bForwardOnly) And pos.pi > 1 Then
      If LTimeSlots(pos.pi - 2).lDate = LTimeSlots(pos.pi).lDate Then
        numConflicts += NumStudentConflicts(ei, pos.pi - 2)
      End If
    End If
    ' After (same day)
    If pos.pi < NumTimeSlots - 2 Then
      If LTimeSlots(pos.pi + 2).lDate = LTimeSlots(pos.pi).lDate Then
        numConflicts += NumStudentConflicts(ei, pos.pi + 2)
      End If
    End If
    If numExamConflicts > 0 Then
      numTwoInADay += numExamConflicts
    End If
    'If numConflicts > 0 Then
    '  numTwoInADay += 1
    'End If
    penDiff = (penTwoInADay * numConflicts)
    totpenTwoInADay += penDiff
    pen += penDiff
    If bReporting And penDiff > 0 Then
      WriteLine("Exam " & ei & ": TwoInADay Violation with " & numConflicts & " students - Penalty " & penDiff)
    End If


    ' Wider Spreads

    numConflicts = 0
    numExamConflicts = 0
    If Not (bForwardOnly) Then
      For i = penWiderSpread To 1 Step -1 ' i > 1; i--)
        If pos.pi - i >= 0 Then
          numConflicts += NumStudentConflicts(ei, pos.pi - i)
        End If
      Next
    End If
    For i = penWiderSpread To 1 Step -1
      If pos.pi + i < NumTimeSlots Then
        numConflicts += NumStudentConflicts(ei, pos.pi + i)
      End If
    Next
    'If numConflicts > 0 Then
    '  numWiderSpreads += 1
    'End If
    If numExamConflicts > 0 Then
      numWiderSpreads += numExamConflicts
    End If
    penDiff = numConflicts
    totpenWiderSpread += penDiff
    pen += penDiff
    If bReporting And penDiff > 0 Then
      WriteLine("Exam " & ei & ": Wider Spread Violation with " & numConflicts & " students - Penalty " & penDiff)
    End If


    ' Penalty on Period
    If pos.pi >= 0 And pos.pi < NumTimeSlots Then
      penDiff = LTimeSlots(pos.pi).penalty
      totpenPeriodPenalty += penDiff
      pen += penDiff
      If penDiff > 0 Then
        numPeriodPenalties += 1
      End If
      If bReporting And penDiff > 0 Then
        WriteLine("Exam " & ei & ": Period penalty Violation - Penalty " & penDiff)
      End If
    End If

    ' Penalty on Room
    If pos.ri >= 0 And pos.ri < NumRooms Then
      penDiff = LRooms(pos.ri).penalty
      totpenRoomPenalty += penDiff
      pen += penDiff
      If penDiff > 0 Then
        numRoomPenalties += 1
      End If
      If bReporting And penDiff > 0 Then
        WriteLine("Exam " & ei & ": Room penalty Violation - Penalty " & penDiff)
      End If
    End If

    ' Penalty on FrontLoad violations
    If pos.pi >= 0 And pos.pi < NumTimeSlots Then
      If LTimeSlots(pos.pi).bLastPeriods And LExams(ei).bLargestExams Then
        penDiff = penFrontLoading
        totpenFrontLoadPenalty += penDiff
        pen += penDiff
        If penDiff > 0 Then
          numFrontLoadPenalties += 1
        End If
        If bReporting And penDiff > 0 Then
          WriteLine("Exam " & ei & ": Front Load Violation at period " & pos.pi & " of " & NumTimeSlots & " (Size " & LExams(ei).ClassSize & ") - Penalty " & penDiff)
        End If
      End If
    End If

    If False Then
      ' Penalty on Different Durations
      If pos.ri >= 0 And pos.ri < NumRooms And pos.pi >= 0 And pos.pi < NumTimeSlots Then

        ' If there are exams (i.e. not first) in this room/timeslot, then check for the exam duration we're adding
        ' If it's already there with another exam, penalty for this room/timeslot will not change or be affected by placing the exam there
        ' Note: May need to check what affect moving the exam from it's original location will have

        If LRooms(pos.ri).Events(pos.pi).Count > 0 Then
          ' At least one exam there - any with same duration?
          Dim dei As Integer, bNoEffect As Boolean = False
          For i = 0 To LRooms(pos.ri).Events(pos.pi).Count - 1
            dei = CInt(CType(LRooms(pos.ri).Events(pos.pi), ArrayList)(i))
            If dei > 0 And dei < NumExams Then
              If LExams(dei).Duration = LExams(ei).Duration Then
                ' Exam with this duration already there - no effect on this penalty
                bNoEffect = True
                Exit For
              End If
            End If
          Next
          If Not (bNoEffect) Then
            ' Some effect
            pen += numMixDurPenalties
          End If
        End If

      End If
    End If

    Return pen
  End Function

  Dim olddiffpen As Long = 0, newdiffpen As Long = 0

  ' Used for Evaluate
  Private Function AbsoluteDiffDurationsPenalty() As Long

    Dim dei As Integer, durs As New ArrayList, i As Integer
    Dim ri As Integer, pi As Integer, penDiff As Long, pen As Long = 0
    Dim diffCount As Integer

    ' Check all events in this room at this time - count how many are unique occurences there are
    ' Result is number of different Durations minus 1 (by weight)

    olddiffpen = newdiffpen

    pen = 0

    Dim durCount As Integer = 0, strDur As String

    For ri = 0 To NumRooms - 1
      For pi = 0 To NumTimeSlots - 1

        '      exCount += LRooms(ri).Events(pi).Count

        ' Collect the durations
        durs.Clear()
        For i = 0 To LRooms(ri).Events(pi).Count - 1
          dei = CInt(CType(LRooms(ri).Events(pi), ArrayList)(i))
          If dei >= 0 And dei < NumExams Then
            durs.Add(LExams(dei).Duration)
          End If
        Next
        If durs.Count > 0 Then
          ' Sort
          durs.Sort()
          ' Then look for differences
          diffCount = 0
          strDur = "" '" of " & durs.Count & " exams: "
          Dim curDur As Integer = CInt(durs(0))
          durCount = 1
          For i = 1 To durs.Count - 1
            If CInt(durs(i)) <> curDur Then
              '         strDur &= curDur & "x" & durCount & " "
              ' If last one, add it too
              'If i = durs.Count - 1 Then
              '  strDur &= CInt(durs(i)) & "x1"
              'End If
              durCount = 0
              diffCount += 1
              curDur = CInt(durs(i))
            End If
            strDur &= CInt(durs(i)) & " "
            durCount += 1
          Next
          ' Apply the penalty
          penDiff = (diffCount * penDiffDurations)
          totpenMixDurPenalty += penDiff
          pen += penDiff
          If penDiff > 0 Then
            numMixDurPenalties += 1
          End If
          If bReporting And penDiff > 0 Then
            WriteLine("Room " & ri & " at Period " & pi & ": Mix Durations penalty Violation - Penalty " & penDiff & " (" & strDur & ")")
          End If
        End If

      Next
    Next

    '    WriteLine(durs.Count & ", " & diffCount & ", " & penDiff & ", " & pen & ", " & exCount)

    newdiffpen = pen

    Return pen

  End Function

  Private Function NumStudentConflicts(ByVal ei As Integer, ByVal ti As Integer) As Integer
    Dim i As Integer, dei As Integer, numConflicts As Integer = 0, sum As Integer = 0
    'For i = 0 To LTimeSlots(ti).events.Count - 1
    '  dei = CInt(LTimeSlots(ti).events(i))
    '  numConflicts += conMat(ei, dei)
    'Next
    'Dim strtext As String
    'For i = 0 To LExams(ei).ConflictExams.Count - 1
    '  strtext &= CType(LExams(ei).ConflictExams(i), IndexValue).Index & " "
    'Next
    'WriteLine("ei=" & ei & " at " & LExams(ei).pos.pi & ":  " & strtext)

    For i = 0 To LExams(ei).ConflictExams.Count - 1
      dei = CType(LExams(ei).ConflictExams(i), IndexValue).Index
      '     WriteLine("ei=" & ei & " at " & LExams(ei).pos.pi & "  dei=" & dei & " checking ti=" & ti & ", Num students involved: " & CType(LExams(ei).ConflictExams(i), IndexValue).Value)
      If LExams(dei).pos.pi = ti Then
        ' Same time as ti, and conflicts - return numStudents involved in conflict
        If CType(LExams(ei).ConflictExams(i), IndexValue).Value > 0 Then
          '    WriteLine("ei=" & ei & " at " & LExams(ei).pos.pi & "  dei=" & dei & " checking ti=" & ti & ", Num students involved: " & CType(LExams(ei).ConflictExams(i), IndexValue).Value)
          numExamConflicts += 1
        End If
        sum += CType(LExams(ei).ConflictExams(i), IndexValue).Value
        '      Return CType(LExams(ei).ConflictExams(i), IndexValue).Value


      End If
    Next
    Return sum
  End Function


  Private Sub Schedule(ByVal ei As Integer, ByVal pos As LPos)
    If ei >= 0 And ei < NumExams Then
      ' Add the pos (ti,ri) to Exam
      LExams(ei).pos.pi = pos.pi
      If pos.pi >= 0 And pos.pi < NumTimeSlots Then
        If (pos.ri < 0 Or pos.ri >= NumRooms) Then
          LExams(ei).pos.ri = FindRoom(ei, pos)
          If LExams(ei).pos.ri = -1 Then
            LExams(ei).pos.pi = -1
            Exit Sub
            Console.WriteLine("Here")
          End If
        Else
          LExams(ei).pos.ri = pos.ri
        End If

        ' Increase the overall room usage at that time
        LRooms(LExams(ei).pos.ri).Usage(pos.pi) += LExams(ei).ClassSize
        ' Add to events in room at this time
        If LRooms(LExams(ei).pos.ri).Events(pos.pi) Is Nothing Then
          LRooms(LExams(ei).pos.ri).Events(pos.pi) = New ArrayList
        End If
        LRooms(LExams(ei).pos.ri).Events(pos.pi).Add(ei)

        '' Calculate the number of mix durations in the room at this time
        'Dim dei As Integer, bNoEffect As Boolean = False, i As Integer
        'For i = 0 To LRooms(pos.ri).Events(pos.pi).Count - 1
        '  dei = CInt(CType(LRooms(pos.ri).Events(pos.pi), ArrayList)(i))
        '  If dei > 0 And dei < NumExams Then
        '    If LExams(dei).Duration = LExams(ei).Duration Then
        '      ' Exam with this duration already there - no effect on this penalty
        '      bNoEffect = True
        '      Exit For
        '    End If
        '  End If
        'Next
        'If Not (bNoEffect) Then
        '  ' Some effect
        '  pen += numMixDurPenalties
        'End If


      End If
    End If
  End Sub

  Private Sub ScheduleRecursive(ByVal pos As LPos, ByVal ei As Integer)
    Schedule(ei, pos)

    ' Setup coincidental pos
    Dim pos2 As New LPos(pos.pi, -1)
    ' Use same room as pos if asked for
    'If LEvent(ei).sameroom Then
    '  pos2.RoomIndex2 = pos.RoomIndex2
    'End If

    If LExams(ei).coincidenceExams.Count > 0 Then
      Dim i As Integer, dei As Integer, sameRoom As Boolean
      For i = 0 To LExams(ei).coincidenceExams.Count - 1
        dei = CInt(LExams(ei).coincidenceExams(i))
        '      sameRoom = CBool(LExams(ei).sameroom(i))
        ' Use same room as pos if asked for
        'If sameRoom Then
        '  pos2.RoomIndex2 = pos.RoomIndex2
        '  pos2.Size = LExams(dei).students.Count
        'Else
        '  pos2.ClearRoomAllocation()
        'End If
        'If LEvent(ei).sameroom Then
        '  pos2.Size = LEvent(dei).students.Count
        'End If
        If (CanSchedule(dei, pos2)) Then
          ' Clear room allocation (which could be set on the CanScheduleEvent) if we don't have a 'sameroom' cooincidence
          'If Not sameRoom Then
          '  pos2.ClearRoomAllocation()
          'End If
          Schedule(dei, pos2)
        End If
      Next
    End If
  End Sub

  Private Sub Unschedule(ByVal ei As Integer)
    Dim pos As LPos = LExams(ei).pos
    If ei >= 0 And ei < NumExams Then
      ' Reduce usage of class / events
      If pos.pi >= 0 And pos.pi < NumTimeSlots And pos.ri >= 0 And pos.ri < NumRooms Then
        ' Usage
        LRooms(pos.ri).Usage(pos.pi) -= LExams(ei).ClassSize
        If LRooms(pos.ri).Usage(pos.pi) < 0 Then
          LRooms(pos.ri).Usage(pos.pi) = 0
          Console.WriteLine("Problem with de-allocation")
          '          MsgBox("Problem with de-allocation")
        End If
        ' Events
        LRooms(pos.ri).Events(pos.pi).Remove(ei)
      End If
      ' Clear the time/room
      LExams(ei).pos = New LPos
    End If
  End Sub

  Private Sub ClearSchedule()
    Dim ei As Integer, ri As Integer, pi As Integer
    For ei = 0 To NumExams - 1
      LExams(ei).pos = New LPos
    Next
    ' Clear Room Usage / Events
    For ri = 0 To NumRooms - 1
      LRooms(ri).Usage = New Integer(NumTimeSlots) {}
      LRooms(ri).Events = New ArrayList(NumTimeSlots) {}
      For pi = 0 To NumTimeSlots - 1
        LRooms(ri).Events(pi) = New ArrayList
      Next
    Next

  End Sub

#End Region

#Region "Problem File Code"

  Private Function LoadData(ByVal filename As String) As Boolean
    Dim path As String = filename & ".exam"
    If Not (File.Exists(path)) Then
      Return False
    End If
    Dim sr As StreamReader = New StreamReader(path)
    LoadDataFrom(sr)
    sr.Close()
    Return True
  End Function

  Public bWebValidator As Boolean = False

  Private Sub WriteLine(Optional ByVal strText As String = "")
    If bWebValidator Then
      lines.Add(strText)
    Else
      Console.WriteLine(strText)
    End If
  End Sub

  Private Sub Write(ByVal strText As String)
    If bWebValidator Then
      lines.Add(strText)
    Else
      Console.Write(strText)
    End If
  End Sub

  Public Sub LoadDataFrom(ByVal sr As Object) 'ByVal filename As String)

    Dim line As String, si As Integer, ei As Integer, resType As String, i As Integer, j As Integer
    Dim capacity As Integer, penalty As Integer, perTime As String, perDuration As Integer
    Dim dur As Integer, maxSi As Integer, maxSize As Integer, maxCapacity As Integer, perDay As String, perMonth As String, perYear As String
    Dim maxDuration As Integer, command As String, dei As Integer, hcCount As Integer = 0, badCount As Integer = 0
    Dim bFound As Boolean

    Dim examDurInfo As New ArrayList  ' of IndexValue
    Dim examSizeInfo As New ArrayList ' of IndexValue

    WriteLine("Reading Problem File")
    WriteLine("====================")

    Do While sr.Peek() >= 0

      line = sr.ReadLine()
      resType = ReadComma(line, , ":")

      If resType = "[Exams" Then
        NumExams = CInt(ReadComma(line, , "]"))
        LExams = New LExam(NumExams) {}
        ExamStudents = New ArrayList(NumExams) {}
        WriteLine("Reading " & NumExams & " Exams... ")
        ' Process the lines
        maxDuration = 0
        For i = 0 To NumExams - 1
          line = sr.ReadLine()
          dur = CInt(ReadComma(line))
          LExams(i) = New LExam(dur)
          ExamStudents(i) = New ArrayList
          ' Read Students
          While line <> ""
            si = Val(ReadComma(line))
            ExamStudents(i).Add(si)
            If si > maxSi Then
              maxSi = si
            End If
          End While

          LExams(i).ClassSize = ExamStudents(i).Count

          If ExamStudents(i).Count > maxSize Then
            maxSize = ExamStudents(i).Count
          End If
          If dur > maxDuration Then
            maxDuration = dur
          End If

          ' Find the exam duration in the list and add one to the count for it
          bFound = False
          For j = 0 To examDurInfo.Count - 1
            If CType(examDurInfo(j), IndexValue).Index = dur Then
              CType(examDurInfo(j), IndexValue).Value += 1
              bFound = True
              Exit For
            End If
          Next
          ' If didn't find it, then add it with a starting count of 1
          If Not (bFound) Then
            examDurInfo.Add(New IndexValue(dur, 1))
          End If
          ' Find the exam size in the list and add one to the count for it
          bFound = False
          For j = 0 To examSizeInfo.Count - 1
            If CType(examSizeInfo(j), IndexValue).Index = Int(LExams(i).ClassSize / 10) Then
              CType(examSizeInfo(j), IndexValue).Value += 1
              bFound = True
              Exit For
            End If
          Next
          ' If didn't find it, then add it with a starting count of 1
          If Not (bFound) Then
            examSizeInfo.Add(New IndexValue(Int(LExams(i).ClassSize / 10), 1))
          End If

          '      Console.WriteLine("Exam " & i & ", dur=" & dur & ", numstu=" & ExamStudents(i).Count)
        Next

        WriteLine("Students: " & CStr(maxSi + 1))

        If bReportOnProblemData Then
          WriteLine("Total Students=" & maxSi + 1 & ", MaxClassSize=" & maxSize & ", MaxDuration=" & maxDuration)
        End If
        NumStudents = maxSi + 1

        If bReportOnProblemData Then
          ' Report on duration and size Profile for exams
          ' Duration
          examDurInfo.Sort(New IndexValue.ivSortMethodIndex)
          For j = 0 To examDurInfo.Count - 1
            With CType(examDurInfo(j), IndexValue)
              Console.WriteLine("Exams with duration " & .Index & ": " & .Value)
            End With
          Next
          ' Size
          examSizeInfo.Sort(New IndexValue.ivSortMethodIndex)
          For j = 0 To examSizeInfo.Count - 1
            With CType(examSizeInfo(j), IndexValue)
              '            Console.WriteLine("Exams with Size " & .Index & ": " & .Value)
              WriteLine("Exams with Size " & .Index * 10 & " to " & (.Index + 1) * 10 & ": " & .Value)
            End With
          Next
        End If

      ElseIf resType = "[Rooms" Then
        ' Read the Number
        NumRooms = CInt(ReadComma(line, , "]"))
        LRooms = New LRoom(NumRooms) {}
        WriteLine("Reading " & NumRooms & " Rooms... ")
        ' Process the lines
        For i = 0 To NumRooms - 1
          line = sr.ReadLine()
          capacity = CInt(ReadComma(line))
          penalty = CInt(ReadComma(line))
          LRooms(i) = New LRoom(capacity, penalty, NumTimeSlots) ' NumTimeSlots will already have been read in so we can dictate this for Room Usage / Period
          If capacity > maxCapacity Then
            maxCapacity = capacity
          End If
          If bReportOnProblemData Then
            WriteLine("Room " & i & ": " & capacity)
          End If
        Next
        If bReportOnProblemData Then
          WriteLine("MaxRoomCapacity=" & maxCapacity)
        End If

      ElseIf resType = "[Periods" Then
        ' Read the Number
        NumTimeSlots = CInt(ReadComma(line, , "]"))
        LTimeSlots = New LTimeSlot(NumTimeSlots) {}
        WriteLine("Reading " & NumTimeSlots & " Timeslots... ")
        ' Process the lines
        maxDuration = 0
        For i = 0 To NumTimeSlots - 1
          line = sr.ReadLine()
          perDay = ReadComma(line, , ":")
          perMonth = ReadComma(line, , ":")
          perYear = ReadComma(line)
          perTime = ReadComma(line) ', , ",")
          perDuration = CInt(ReadComma(line))
          penalty = CInt(ReadComma(line))

          LTimeSlots(i) = New LTimeSlot(TimeSpan.Parse(perTime), New Date(Val(perYear), Val(perMonth), Val(perDay)), perDuration, penalty)
          If bReportOnProblemData Then
            WriteLine("Period " & i & ": " & LTimeSlots(i).lDate.ToShortDateString & ", " & LTimeSlots(i).Time.ToString & ", " & perDuration)
          End If

          If perDuration > maxDuration Then
            maxDuration = perDuration
          End If
        Next
        If bReportOnProblemData Then
          WriteLine("Max Duration= " & maxDuration)
        End If

      ElseIf resType = "[PeriodHardConstraints]" Then


        ' Read Period Hard Constraints
        line = ""
        Do While sr.Peek() >= 0 And line <> "[RoomHardConstraints]"
          '  lines.Add("[" & Me.line & "]")
          line = sr.ReadLine()


          If line <> "[RoomHardConstraints]" Then

            ' Process the different constraints
            ei = Val(ReadComma(line))
            If ei >= 0 And ei < NumExams Then

              command = Trim(ReadComma(line))

              ' Commands
              If command = "EXAM_COINCIDENCE" Then
                dei = Val(ReadComma(line))
                If dei >= 0 And dei < NumExams Then
                  LExams(ei).coincidenceExams.Add(dei)
                  LExams(dei).coincidenceExams.Add(ei)
                  hcCount += 1
                Else
                  WriteLine("Bad exam index (" & dei & "): " & line)
                  badCount += 1
                End If
              ElseIf command = "EXCLUSION" Then
                dei = Val(ReadComma(line))
                If dei >= 0 And dei < NumExams Then
                  LExams(ei).excludesExams.Add(dei)
                  LExams(dei).excludesExams.Add(ei)
                  hcCount += 1
                Else
                  WriteLine("Bad exam index (" & dei & "): " & line)
                  badCount += 1
                End If
                'ElseIf command = "PRECEDES" Then
                '  dei = Val(ReadComma(line))
                '  If dei >= 0 And dei < NumExams Then
                '    LExams(ei).afterExams.Add(dei)
                '    hcCount += 1
                '  Else
                '    WriteLine("Bad exam index (" & dei & "): " & line)
                '    badCount += 1
                '  End If
              ElseIf command = "AFTER" Then '"FOLLOWS" Then
                dei = Val(ReadComma(line))
                If dei >= 0 And dei < NumExams Then
                  LExams(ei).beforeExams.Add(dei)
                  LExams(dei).afterExams.Add(ei)
                  hcCount += 1
                Else
                  WriteLine("Bad exam index (" & dei & "): " & line)
                  badCount += 1
                End If
              Else
                WriteLine("Unknown Constratint: " & line)
                badCount += 1
              End If

            Else
              WriteLine("Bad exam index (" & ei & "): " & line)
              badCount += 1
            End If

          End If
        Loop

        ' Read Room Hard Constraints
        line = ""
        Do While sr.Peek() >= 0 And line <> "[InstitutionalWeightings]"
          '  lines.Add("[" & Me.line & "]")
          line = sr.ReadLine()


          If line <> "[InstitutionalWeightings]" Then

            ' Process the different constraints
            ei = Val(ReadComma(line))
            If ei >= 0 And ei < NumExams Then

              command = Trim(ReadComma(line))

              ' Commands
              If command = "ROOM_EXCLUSIVE" Then
                LExams(ei).bExclusiveRoom = True
                hcCount += 1
              Else
                WriteLine("Unknown Constratint: " & line)
                badCount += 1
              End If

            Else
              WriteLine("Bad exam index (" & ei & "): " & line)
              badCount += 1
            End If

          End If
        Loop


        Write("Valid Hard Constraints: " & hcCount)
        If badCount > 0 Then
          WriteLine("Bad Hard Constraints: " & badCount)
        Else
          WriteLine(" (All okay)")
        End If

        ' Now read Soft Constraints
        Do While sr.Peek() >= 0
          line = sr.ReadLine()
          resType = ReadComma(line)
          If resType = "TWOINAROW" Then
            penTwoInARow = CInt(ReadComma(line))
          ElseIf resType = "TWOINADAY" Then
            penTwoInADay = CInt(ReadComma(line))
          ElseIf resType = "PERIODSPREAD" Then
            penWiderSpread = CInt(ReadComma(line))
          ElseIf resType = "NOMIXEDDURATIONS" Then
            penDiffDurations = Val(ReadComma(line))
          ElseIf resType = "FRONTLOAD" Then
            frontNumLargestExams = Val(ReadComma(line))
            frontNumLastPeriods = Val(ReadComma(line))
            penFrontLoading = Val(ReadComma(line))
          End If
        Loop

        WriteLine("Institutional Weightings: TwoInARow=" & penTwoInARow & ", TwoInADay=" & penTwoInADay & ", PeriodSpread=" & penWiderSpread)
        WriteLine("Institutional Weightings: NoMixedDurations=" & penDiffDurations & ", FrontLoad - NumberOfLargestExams=" & frontNumLargestExams & _
        ", FrontLoad - NumberOfLastPeriodsAvoid=" & frontNumLastPeriods & ", FrontLoad - Penalty=" & penFrontLoading)

      End If

    Loop

    If True Then
      ' Now do some checks - first, report any exams which are too big for the available rooms
      For ei = 0 To NumExams - 1
        If LExams(ei).ClassSize > maxCapacity Then
          WriteLine("Exam " & ei & " (" & LExams(ei).ClassSize & ") is too big - Max Room Capacity is " & maxCapacity)
          ' Remove end students
          For i = ExamStudents(ei).Count - 1 To maxCapacity - 1 Step -1
            ExamStudents(ei).RemoveAt(i)
          Next
          LExams(ei).ClassSize = maxCapacity

        End If
      Next
      If False Then
        For ei = 0 To NumExams - 1
          If LExams(ei).ClassSize > maxCapacity Then
            WriteLine("Exam " & ei & " (" & LExams(ei).ClassSize & ") is too big - Max Room Capacity is " & maxCapacity)
            ' Remove end students
            For i = ExamStudents(ei).Count - 1 To maxCapacity - 1 Step -1
              ExamStudents(ei).RemoveAt(i)
            Next
            LExams(ei).ClassSize = maxCapacity

          End If
        Next
      End If
    End If

    ' Build up some clash and resource info
    BuildConflicts()
    BuildAllocationLists()

    ' Now check for those exams which are down as cooincidences, but have conflicts
    For ei = 0 To NumExams - 1
      For i = 0 To LExams(ei).coincidenceExams.Count - 1
        dei = CInt(LExams(ei).coincidenceExams(i))
        ' Does this exam also reside in conflicts?
        bFound = False
        For j = 0 To LExams(ei).ConflictExams.Count - 1
          If CType(LExams(ei).ConflictExams(j), IndexValue).Index = dei Then
            bFound = True
            Exit For
          End If
        Next
        If bFound Then
          WriteLine("Exam " & ei & " requires cooincidence (same time) as exam " & dei & " but they conflict")
        End If
      Next
    Next

  End Sub

  Private Sub BuildConflicts()

    Dim si As Integer, ei As Integer, dei As Integer, i As Integer, j As Integer, noexamsCount As Integer = 0
    Dim stuExams(NumStudents) As ArrayList

    If bReportOnProblemData Then
      WriteLine("Building Conflict Matrix - First Phase")
    End If

    ' First pass - for each student, collect the exams they're doing
    For ei = 0 To NumExams - 1
      LExams(ei).ConflictExams.Clear()
      For i = 0 To ExamStudents(ei).Count - 1
        si = CInt(CType(ExamStudents(ei), ArrayList)(i))
        If stuExams(si) Is Nothing Then
          stuExams(si) = New ArrayList
        End If
        stuExams(si).Add(ei)
      Next
    Next

    If bReportOnProblemData Then
      WriteLine("Building Conflict Matrix - Second Phase")
    End If

    ' Now work out the exams which must be kept apart (ie. those with a student in common)
    For si = 0 To NumStudents - 1
      If Not (stuExams(si) Is Nothing) Then
        For i = 0 To stuExams(si).Count - 1
          ei = CInt(stuExams(si)(i))
          For j = 0 To stuExams(si).Count - 1
            dei = CInt(stuExams(si)(j))
            If ei <> dei Then
              LExams(ei).ConflictExams.Add(New IndexValue(dei, 1))
            End If
          Next
        Next
      Else
        noexamsCount += 1
      End If
    Next

    If bReportOnProblemData Then
      If noexamsCount > 0 Then
        WriteLine(noexamsCount & " Students have no exams")
      End If
      WriteLine("Building Conflict Matrix - Tidying")
    End If

    ConflictSortAndTidy()

    If bReportOnProblemData Then
      WriteLine("Building Conflict Matrix - Done")
    End If

  End Sub

  Private Sub ConflictSortAndTidy()
    ' Sort the conflict events
    Dim lastei As Integer, copyList As ArrayList, ei As Integer, dei As Integer, i As Integer, dupeCount As Integer = 0, uniqueCount As Integer = 0
    For ei = 0 To NumExams - 1
      ' Sort
      LExams(ei).ConflictExams.Sort(New IndexValue.ivSortMethodIndex)
      ' Remove duplicates
      lastei = -1
      copyList = New ArrayList
      For i = 0 To LExams(ei).ConflictExams.Count - 1
        dei = CType(LExams(ei).ConflictExams(i), IndexValue).Index
        If lastei <> dei Then
          ' First student conflict between ei and dei
          copyList.Add(New IndexValue(dei, 1))
          uniqueCount += 1
        Else
          ' Count subsequent student conflicts between ei and dei
          CType(copyList(copyList.Count - 1), IndexValue).Value += 1
          dupeCount += 1
        End If
        lastei = dei
      Next
      LExams(ei).ConflictExams = copyList
    Next
    If dupeCount > 0 And bReportOnProblemData Then
      WriteLine(dupeCount & " Duplicate exams removed, " & uniqueCount & " exams remaining")
    End If
  End Sub

  Private Sub BuildAllocationLists()
    Dim ei As Integer, pi As Integer, ri As Integer
    For ei = 0 To NumExams - 1
      With LExams(ei)
        ' Periods
        .SuitablePeriods.Clear()
        For pi = 0 To NumTimeSlots - 1
          If .Duration <= LTimeSlots(pi).Duration Then
            .SuitablePeriods.Add(pi)
          End If
        Next
        If .SuitablePeriods.Count = 0 Then
          WriteLine("Exam " & ei & " has no suitable timeslots")
        End If
        ' Rooms
        .SuitableRooms.Clear()
        For ri = 0 To NumRooms - 1
          If .ClassSize <= LRooms(ri).Capacity Then
            .SuitableRooms.Add(ri)
          End If
        Next
        If .SuitableRooms.Count = 0 Then
          WriteLine("Exam " & ei & " has no suitable Rooms")
        End If
      End With
    Next
  End Sub

#End Region

#Region "Solution File Code"

  Private Function LoadSolution(ByVal filename As String) As Boolean
    Dim path As String = filename & ".sln"
    If Not (File.Exists(path)) Then
      Return False
    End If
    Dim sr As StreamReader = New StreamReader(path)
    LoadSolutionFrom(sr)
    sr.Close()
    Return True
  End Function

  Public Sub LoadSolutionFrom(ByVal sr As Object)

    '    Dim path As String = filename & ".sln"
    '   Dim sr As StreamReader = New StreamReader(path)

    Dim line As String, pi As Integer, ri As Integer, linecopy As String, piStr As String, riStr As String, ei As Integer = 0, pos As LPos
    Dim schedExams As Integer = 0, prefix As String

    WriteLine()
    WriteLine("Reading Solution File")
    WriteLine("=====================")

    Do While sr.Peek() >= 0 And ei < NumExams
      line = sr.ReadLine()
      linecopy = line

      prefix = "Exam " & CStr(ei) & ": "

      If line <> "" Then
        piStr = ReadComma(line)
        If IsNumeric(piStr) Then
          pi = CInt(piStr)
          If pi < 0 Or pi >= NumTimeSlots Then
            WriteLine(prefix & "       Invalid Period (" & pi & "): " & linecopy & " - No Schedule")
          Else
            riStr = ReadComma(line)
            If IsNumeric(riStr) Then
              ri = CInt(riStr)
              If ri < 0 Or ri >= NumRooms Then
                WriteLine(prefix & "       Invalid Room (" & ri & "): " & linecopy & " - No Schedule")
              Else
                ' All seems correct - Now can this be scheduled?
                pos = New LPos(pi, ri)
                If CanSchedule(ei, pos) Then
                  'Console.WriteLine("Schedule Okay: Period " & pi & ", Room " & ri)
                  Schedule(ei, pos)
                  schedExams += 1
                Else
                  WriteLine(prefix & "Can't Schedule: Period " & pi & ", Room " & ri)
                End If
              End If
            Else
              WriteLine(prefix & "       Non-Numeric Room (" & riStr & "): " & linecopy & " - No Schedule")
            End If
          End If
        Else
          WriteLine(prefix & "       Non-Numeric Period (" & piStr & "): " & linecopy & " - No Schedule")
        End If
      Else
        WriteLine(prefix & "       Blank Line - No Schedule")
      End If

      ei += 1

    Loop

    '  sr.Close()

    'Console.WriteLine()
    'If schedExams < NumExams Then
    '  Console.WriteLine("Non-feasible Schedule: " & schedExams & " of " & NumExams & " Scheduled (" & NumExams - schedExams & " Unplaced)")
    'End If

    'Console.WriteLine()
    'Console.WriteLine("Evaluation Penalty: " & Evaluate())

    '  If bReportOnProblemData Then
    '    Write("Conflict Density: ")
    Dim cf As Double = 0
    For ei = 0 To NumExams - 1
      cf += LExams(ei).ConflictExams.Count
    Next
    WriteLine("Conflict Density: " & CStr(Math.Round((cf / (NumExams * NumExams)) * 100, 2)) & " %")
    '   End If

  End Sub

  Private Sub SaveSolution(ByVal filename As String)

    Dim path As String = filename & ".sln"
    Dim sw As StreamWriter = New StreamWriter(path)
    Dim ei As Integer

    For ei = 0 To NumExams - 1
      sw.WriteLine(CStr(LExams(ei).pos.pi) & ", " & CStr(LExams(ei).pos.ri))
    Next

    sw.Close()

  End Sub

#End Region

#Region "Common Stuff"

  Public Function Peek() As Integer
    If line <> "" Then
      Return 0
    Else
      Return -1
    End If
  End Function

  Public line As String, lines As New ArrayList

  Public Function ReadLine() As String
    Dim pos As Integer = InStr(line, Chr(10)), retStr As String = ""
    If pos > 0 And pos <= Len(line) Then
      retStr = line.Substring(0, pos - 1 - 1) ' extra reduction for the Chr(13) character
      ' Remove processed text from line
      line = line.Substring(pos, Len(line) - (pos))  ' extra reduction for the Chr(13) character
    Else
      retStr = line
      line = ""
    End If
    '   lines.Add("[" & retStr & "]")
    Return retStr
  End Function

  Public Function ReadComma(ByRef line As String, Optional ByVal bRemoveQuotes As Boolean = True, Optional ByVal searchStr As String = ",", Optional ByVal IgnoreFirstComma As Boolean = False) As String
    ' Read first comma (or other specified) -separated word from line and return, also returning changed line
    ' This removes any quotes if not told otherwise
    If bRemoveQuotes And Len(line) > 0 Then
      If line.Chars(0) = """" Then
        line = line.Substring(1, Len(line) - 1)
      End If
    End If

    Dim pos As Integer = InStr(line, searchStr), retStr As String = ""
    If pos > 0 And pos <= Len(line) Then
      ' May need to skip first occurance
      ' Check whether there's a 'Slash' just before the comma - in which case we ignore and move on - only does this check once, so multiple occurances will be a problem
      '   If IgnoreFirstComma Then
      If pos > 1 Then
        If line.Chars(pos - 2) = "\" Then
          ' Search again, after first occurance
          Dim templine As String = line.Substring(pos, Len(line) - pos)
          Dim temppos = InStr(templine, searchStr)
          ' May not be a further comma - end of line - in which case, just add on to end of line
          If temppos = 0 Then
            temppos = Len(templine)
          End If
          pos += temppos '+ 1
        End If
      End If
      retStr = line.Substring(0, pos - 1)
      ' Remove processed text from line
      line = line.Substring(pos, Len(line) - pos)
    Else
      retStr = line
      line = ""
    End If
    ' Remove leading quote (if exists)
    If retStr <> "" Then
      If retStr.Chars(0) = """" Then
        retStr = retStr.Substring(1, Len(retStr) - 1)
      End If
    End If
    ' Remove trailing quote (if exists)
    If retStr <> "" Then
      If retStr.Chars(Len(retStr) - 1) = """" Then
        retStr = retStr.Substring(0, Len(retStr) - 1)
      End If
    End If
    Return retStr
  End Function

#End Region

End Class
